###########################################################################
#
#  Script to utilise an external program to retrieve scores/teams on
# Q3F servers, check if score has changed, and report if it has.
#
# Copyright 2001,2002 Neil Peter Charley (aka Athan on QuakeNet),
# all rights reserved
#
# Contact: q3f-score@gurus.tf
#
# Version:
set q3fscore_version "0.07"
#
# TODO:
# - command "listservers" all info for each
# - q3f-commands command - summary of all commands
#
# See end of file for changelog
###########################################################################

###########################################################################
# We don't want no stinking locales
#  If someone has a broken/incomplete locale, but set, it may confuse perl
#  which will then spit out nasty error messages which _may_ confuse our
#  parseing of q3f-score's output
###########################################################################
global env
set env(LANGUAGE) ""
set env(LC_ALL) ""
set env(LC_MESSAGES) ""
set env(LC_TIME) ""
set env(LC_NUMERIC) ""
set env(LC_CTYPE) ""
set env(LC_MONETARY) ""
set env(LC_COLLATE) ""
set env(LANG) ""
###########################################################################

###########################################################################
# Global variables
###########################################################################
set default_period 30
set min_period 5
set default_channel "#athan"
set default_matchlength 30
set noteamscore -9999
###########################################################################

###########################################################################
# Triggers
###########################################################################
bind msg o|o addserver do_addserver
bind msg o|o delserver do_delserver
bind msg o|o report do_reportserver
bind msg o|o period do_changeperiod
bind msg o|o length do_changematchlength
bind msg o|o noend do_noend
bind msg o|o channel do_changechannel
bind msg o|o status do_status
bind msg o|o timers do_timers
###########################################################################

###########################################################################
# Procedures
###########################################################################
### Process request to add a server to the monitored list
proc do_addserver { nick uhost handle arg } {
	global q3f_servers default_period default_channel default_matchlength noteamscore

	if { ![matchattr $handle o] && ![matchattr $handle |o $default_channel] } {
		puthelp "NOTICE $nick :q3f-score: You need to be a global op or op in $default_channel to set up score reporting, sorry."
		return 1
	}
	set args [split $arg]
	set server [lindex $args 0]
	set channel [lindex $args 1]
	set matchlength [lindex $args 2]
	set period [lindex $args 3]
	set noend [lindex $args 4]
	if { [string match "" $server] } {
		putcmdlog "do_addserver: No server arg supplied"
		puthelp "NOTICE $nick :Format: addserver <host:port> [<#channel> [<matchlength> [<query period> [<noend>]]]]"
		return 1
	}
	if { ! [string match "" [array get q3f_servers "$server,team1"]] } {
		putcmdlog "do_addserver: server already set"
		puthelp "NOTICE $nick :q3f-score: Server '$server' already set"
		return 1
	}
	set q3f_servers($server,team1) "Red"
	set q3f_servers($server,team2) "Blue"
	set q3f_servers($server,team3) "Yellow"
	set q3f_servers($server,team4) "Green"
	set q3f_servers($server,teamscore1) $noteamscore
	set q3f_servers($server,teamscore2) $noteamscore
	set q3f_servers($server,teamscore3) $noteamscore
	set q3f_servers($server,teamscore4) $noteamscore
	set q3f_servers($server,timerid) -1
	if { [string match "" $period] } {
		set q3f_servers($server,period) $default_period
	} else {
		set q3f_servers($server,period) $period
	}
	if { [string match "" $matchlength] } {
		set q3f_servers($server,matchlength) $default_matchlength
	} else {
		set q3f_servers($server,matchlength) $matchlength
	}
	if { [string match "" $channel] } {
		set q3f_servers($server,channel) $default_channel
	} else {
		set q3f_servers($server,channel) $channel
	}
	if { [string match "" $noend] } {
		set q3f_servers($server,noend) 0
	} else {
		set q3f_servers($server,noend) $noend
	}
	set q3f_servers($server,started) 0
	putcmdlog "do_addserver: Server '$server' added to monitoring"
	puthelp "NOTICE $nick :q3f-score: Server '$server' added to monitoring ($q3f_servers($server,channel), len: $q3f_servers($server,matchlength)mins, period: $q3f_servers($server,period)secs)"

	# Kick off checking of server
	check_server $server

	return 0
}

### Process request to delete a server from the monitored list
proc do_delserver { nick uhost handle arg } {
	global q3f_servers

	if { [string match "" $arg] } {
		putcmdlog "do_delserver: No server arg supplied"
		puthelp "NOTICE $nick :Format: delserver <host:port>"
		return 1
	}
	if { [string match "" [array get q3f_servers "$arg,team1"]] } {
		putcmdlog "do_delserver: server already deleted"
		puthelp "NOTICE $nick :q3f-score: Server already deleted."
		return 1
	}
	if { ![matchattr $handle o] && ![matchattr $handle |o $q3f_servers($arg,channel)] } {
		puthelp "NOTICE $nick :q3f-score: You need to be a global op or op in $q3f_servers($arg,channel) to remove this reporting, sorry."
		return 1
	}
	unset q3f_servers($arg,team1)
	if { $q3f_servers($arg,timerid) != -1 } {
		killutimer $q3f_servers($arg,timerid)
		set q3f_servers($arg,timerid) -1
	}
	putcmdlog "do_delserver: server deleted."
	puthelp "NOTICE $nick :q3f-score: server deleted ($arg)"

	return 0;
}

## Force a report of the given server
proc do_reportserver { nick uhost handle arg } {
	global q3f_servers

	if { ![matchattr $handle o] && ![matchattr $handle |o $default_channel] } {
		puthelp "NOTICE $nick :q3f-score: You need to be a global op or op in $default_channel to set up score reporting, sorry."
		return 1
	}
	if { [string match "" $arg] } {
		putcmdlog "do_reportserver: No server arg supplied"
		puthelp "NOTICE $nick :Format: report <host:port>"
		return 1
	}
	if { [string match "" [array get q3f_servers "$arg,team1"]] } {
		putcmdlog "do_reportserver: server not being monitored."
		puthelp "NOTICE $nick :q3f-score: Server not being monitored."
		return 1
	}
	if { ![matchattr $handle o] && ![matchattr $handle |o $q3f_servers($arg,channel)] } {
		puthelp "NOTICE $nick :q3f-score: You need to be a global op or op in $q3f_servers($arg,channel) to force a report, sorry."
		return 1
	}
	report_score $arg
}

## Change the poll period of the given server
proc do_changeperiod { nick uhost handle arg } {
	global q3f_servers min_period

	set args [split $arg " "]
	set server [lindex $args 0]
	set period [lindex $args 1]
	if { [string match "" $arg] || [string match "" $period] } {
		putcmdlog "do_changeperiod: Incorrect arguments"
		puthelp "NOTICE $nick :Format: period <host:port> <seconds>"
		return 1
	}
	if { [string match "" [array get q3f_servers "$server,team1"]] } {
		putcmdlog "do_changeperiod: server not being monitored."
		puthelp "NOTICE $nick :q3f-score: Server not being monitored."
		return 1
	}
	if { ![matchattr $handle o] && ![matchattr $handle |o $q3f_servers($server,channel)] } {
		puthelp "NOTICE $nick :q3f-score: You need to be a global op or op in $q3f_servers($server,channel) to change the polling period, sorry."
		return 1
	}

	if { $period < $min_period } {
		puthelp "NOTICE $nick :q3f-score: Minimum poll period is $min_period, try again."
		return 1
	}
	set q3f_servers($server,period) $period
	puthelp "NOTICE $nick :q3f-score: Poll period for $server set to $period seconds."

	return 0
}

## Change the match length of the given server
proc do_changematchlength { nick uhost handle arg } {
	global q3f_servers

	set args [split $arg " "]
	set server [lindex $args 0]
	set matchlength [lindex $args 1]
	if { [string match "" $arg] || [string match "" $matchlength] } {
		putcmdlog "do_changematchlength: Incorrect arguments"
		puthelp "NOTICE $nick :Format: period <host:port> <minutes>"
		return 1
	}
	if { [string match "" [array get q3f_servers "$server,team1"]] } {
		putcmdlog "do_changematchlength: server not being monitored."
		puthelp "NOTICE $nick :q3f-score: Server not being monitored."
		return 1
	}
	if { ![matchattr $handle o] && ![matchattr $handle |o $q3f_servers($server,channel)] } {
		puthelp "NOTICE $nick :q3f-score: You need to be a global op or op in $q3f_servers($server,channel) to change the match length, sorry."
		return 1
	}

	set q3f_servers($server,matchlength) $matchlength
	puthelp "NOTICE $nick :q3f-score: Match Length for $server set to $matchlength minutes."

	return 0
}

## Change the 'noend' variable.  Some maps MAY have teamscores go backwards (Castles ?)
proc do_noend { nick uhost handle arg } {
	global q3f_servers

	set args [split $arg " "]
	set server [lindex $args 0]
	set noend [lindex $args 1]
	if { [string match "" $arg] || [string match "" $noend] } {
		putcmdlog "do_noend: Incorrect arguments"
		puthelp "NOTICE $nick :Format: noend <host:port> 0|1"
		return 1
	}
	if { [string match "" [array get q3f_servers "$server,team1"]] } {
		putcmdlog "do_noend: server not being monitored."
		puthelp "NOTICE $nick :q3f-score: Server not being monitored."
		return 1
	}
	if { ![matchattr $handle o] && ![matchattr $handle |o $q3f_servers($server,channel)] } {
		puthelp "NOTICE $nick :q3f-score: You need to be a global op or op in $q3f_servers($server,channel) to changed the 'noend' setting, sorry."
		return 1
	}

	if { ![string match "0" $noend] && ![string match "1" $noend] } {
		putcmdlog "do_noend: Incorrect arguments"
		puthelp "NOTICE $nick :Format: noend <host:port> 0|1"
		return 1
	}
	set q3f_servers($server,noend) $noend
	puthelp "NOTICE $nick :q3f-score: No-end for $server set to $noend."

	return 0
}

## Change the Channel that a server is reported to
proc do_changechannel { nick uhost handle arg } {
	global q3f_servers

	set args [split $arg " "]
	set server [lindex $args 0]
	set channel [lindex $args 1]
	if { [string match "" $arg] || [string match "" $channel] } {
		putcmdlog "do_changechannel: Incorrect arguments"
		puthelp "NOTICE $nick :Format: channel <host:port> <#channelname>"
		return 1
	}
	if { [string match "" [array get q3f_servers "$server,team1"]] } {
		putcmdlog "do_changechannel: server not being monitored."
		puthelp "NOTICE $nick :q3f-score: Server not being monitored."
		return 1
	}
	if { ![matchattr $handle o] && (![matchattr $handle |o $q3f_servers($server,channel)] || ![matchattr $handle |o $channel]) } {
		puthelp "NOTICE $nick :q3f-score: You need to be a global op or op in $q3f_servers($server,channel) AND $channel to change reporting from $q3f_servers($server,channel) to $channel, sorry."
		return 1
	}

	if { ![string match "#*" $channel] } {
		putcmdlog "do_changechannel: Incorrect arguments"
		puthelp "NOTICE $nick :Format: channel <host:port> <#channelname>"
		return 1
	}

	## Is it a channel we're actually on ?
	if { ! [botonchan $channel] } {
		puthelp "NOTICE $nick :q3f-score: Sorry I'm not on $channel so can't report to there."
		return 1
	}
	set q3f_servers($server,channel) $channel
	puthelp "NOTICE $nick :q3f-score: Now reporting to $channel for $server."
	report_score $server

	return 0
}

## Report the status of polling on the given server
proc do_status { nick uhost handle arg } {
	global q3f_servers

	if { [string match "" $arg] } {
		putcmdlog "do_status: Incorrect arguments"
		puthelp "NOTICE $nick :Format: status <host:port>"
		return 1
	}
	if { [string match "" [array get q3f_servers "$arg,team1"]] } {
		putcmdlog "do_changematchlength: server not being monitored."
		puthelp "NOTICE $nick :q3f-score: Server not being monitored."
		return 1
	}

	set status ":q3f-score: $arg being reported to $q3f_servers($arg,channel) every $q3f_servers($arg,period) seconds with match length $q3f_servers($arg,matchlength) minutes."
	if { $q3f_servers($arg,started) != 0 } {
		set length [expr [clock seconds] - $q3f_servers($arg,started)]
		set togo [expr $q3f_servers($arg,matchlength) * 60 - ([clock seconds] - $q3f_servers($arg,started))]
		append status "  Match started approx. " [clock format $length -format %M:%S] " ago, and has " [clock format $togo -format %M:%S] " to go."
	}
	puthelp "NOTICE $nick $status"

	return 0
}

### Actually check a server and schedule next check
proc get_scores { server } {
	global q3f_servers

	set q3fscore [open "|q3f-score $server"]
	if { $q3fscore < 0 } {
		putcmdlog "get_score: q3fscore < 0"
		return 1
	}
	gets $q3fscore scoreline
	while { ! [string match $scoreline "END" ] } {
		if { [string match $scoreline "SCORES" ] } {
			gets $q3fscore scoreline
			set teamno 1
			while { ! [string match $scoreline "END" ] } {
				# Line of form "team-name\tscore"
				set score [split $scoreline "\t"]
				set q3f_servers($server,team$teamno) [lindex $score 0]
				set q3f_servers($server,teamscore$teamno) [lindex $score 1]
				incr teamno
				gets $q3fscore scoreline
			}
			break
		} elseif { $q3f_servers($server,started) == 0 && [string match $scoreline "Match started"] } {
			## q3f-score perl script detected match start
			set q3f_servers($server,started) [clock seconds]
			putcmdlog "q3f-score: match started on $server"
		} elseif { $q3f_servers($server,started) != 0 && [string match $scoreline "No-one playing"] } {
			## Match ended
			close $q3fscore
			unset q3f_servers($server,team1)
			set q3f_servers($server,timerid) -1
			report_score $server
			return 1
		}
		gets $q3fscore scoreline
	}
	close $q3fscore

	if { $q3f_servers($server,started) == 0 } {
		set total [expr $q3f_servers($server,teamscore1) + $q3f_servers($server,teamscore2) + $q3f_servers($server,teamscore3) + $q3f_servers($server,teamscore4)]
		if { $total > 0 } {
			# Teamscores are more than 0, so match must have started
			set q3f_servers($server,started) [clock seconds]
			putcmdlog "q3f-score: match started on $server"
		}
	}
	return 0
}

proc report_score { server } {
	global q3f_servers noteamscore

	set scoreline ""
	if { $q3f_servers($server,teamscore1) != $noteamscore } {
		append scoreline ", " $q3f_servers($server,team1) ": " $q3f_servers($server,teamscore1)
	}
	if { $q3f_servers($server,teamscore2) != $noteamscore } {
		append scoreline ", " $q3f_servers($server,team2) ": " $q3f_servers($server,teamscore2)
	}
	if { $q3f_servers($server,teamscore3) != $noteamscore } {
		append scoreline ", " $q3f_servers($server,team3) ": " $q3f_servers($server,teamscore3)
	}
	if { $q3f_servers($server,teamscore4) != $noteamscore } {
		append scoreline ", " $q3f_servers($server,team4) ": " $q3f_servers($server,teamscore4)
	}
	if { $q3f_servers($server,started) != 0 } {
#		putcmdlog "q3f-score: $server started at $q3f_servers($server,started)"
		set scoretime [expr ($q3f_servers($server,matchlength) * 60) - ([clock seconds] - $q3f_servers($server,started))]
		set scorestr [clock format $scoretime -format %M:%S]
		append scoreline " (" $scorestr ")"
	}
	if { [string length $scoreline] > 0} {
		putcmdlog "q3f-score: $server score now$scoreline"
		puthelp "PRIVMSG $q3f_servers($server,channel) :$server score now$scoreline"
	}

	return 0
}

proc check_server { server } {
	global q3f_servers

	set last_teamscore1 $q3f_servers($server,teamscore1)
	set last_team1 $q3f_servers($server,team1)
	set last_teamscore2 $q3f_servers($server,teamscore2)
	set last_team2 $q3f_servers($server,team2)
	set last_teamscore3 $q3f_servers($server,teamscore3)
	set last_team3 $q3f_servers($server,team3)
	set last_teamscore4 $q3f_servers($server,teamscore4)
	set last_team4 $q3f_servers($server,team4)
	if { [get_scores $server] } {
		## Match ended
		puthelp "PRIVMSG $q3f_servers($server,channel) :Match ended, removing from reporting"
		return 0
	}
	if { ($last_teamscore1 > $q3f_servers($server,teamscore1)
		|| $last_teamscore2 > $q3f_servers($server,teamscore2)
		|| $last_teamscore3 > $q3f_servers($server,teamscore3)
		|| $last_teamscore4 > $q3f_servers($server,teamscore4)) } {
		if { $q3f_servers($server,noend) == 0 } {
			## Match must have ended as score has gone backwards
			report_score $server
			puthelp "PRIVMSG $q3f_servers($server,channel) :Match ended, removing from reporting"
			unset q3f_servers($server,team1)
			set q3f_servers($server,timerid) -1
			return 0
		}
	} elseif { $last_teamscore1 != $q3f_servers($server,teamscore1) 
			|| ![string equal $last_team1 $q3f_servers($server,team1)]
			|| $last_teamscore2 != $q3f_servers($server,teamscore2)
			|| ![string equal $last_team2 $q3f_servers($server,team2)]
			|| $last_teamscore3 != $q3f_servers($server,teamscore3)
			|| ![string equal $last_team3 $q3f_servers($server,team3)]
			|| $last_teamscore4 != $q3f_servers($server,teamscore4)
			|| ![string equal $last_team4 $q3f_servers($server,team4)] } {
		report_score $server
	} else {
#		putcmdlog "Run for $server finished"
	}

	set q3f_servers($server,timerid) [utimer $q3f_servers($server,period) [split "check_server $server"]]
	return 0
}

proc do_timers { nick uhost handle arg } {
		putcmdlog "Timers: [timers]"

		return 0
}
###########################################################################

putlog "Q3F-Score v$q3fscore_version loaded"

## Changleog
# v0.07
#  - Extend addserver command to allow setting of report channel, match length, query period and noend flag all in one go.
#
# v0.06
#  - Corrected error in check_server "[13:43] can't read "arg": no such variable" on match end detection.
#
# v0.05
#  - Kill locale settings so they don't confuse our parseing
#
# v0.04
#  - Fixed 'status' to use period, not matchlength when reporting polling period.
# v0.03
#  - Put user attributes checks on all commands (except status).
#  - Fixed small bug in report_score so we don't try to use a non-existent $scoreline.
# v0.02
#  - Made several things tweakable; channel to report to, period of polling (minimum settable), matchlength, matchend detection
#  - Change in reported team name will trigger a report to channel, not just change in score
#  - Report time left in match on score reports
#  - status command to show full status of a server
